/***************************************************************************/
/*                                                                         */
/*  File: Hair.h                                                           */
/*  Date: 24-10-07                                                         */
/*                                                                         */
/***************************************************************************/
/*
*/
/***************************************************************************/
#ifndef HAIR_H
#define HAIR_H



#include <d3dx9.h>				// Direct X Header Files

								// Direct X Library Files
#pragma comment(lib, "d3d9.lib")	
#pragma comment(lib, "d3dx9.lib")

// DirectX Device/Render Handles
extern LPDIRECT3D9			g_pD3D;			// main.cpp
extern LPDIRECT3DDEVICE9	g_pD3DDevice;	// main.cpp


#include "Maths.h"
#include "dxdbg.h"

// Some helper containers
struct Constraint
{
	int		index0;
	int		index1;
	float	restLength;
};

struct Point
{
	D3DXVECTOR3 curPos;
	D3DXVECTOR3 oldPos;
	bool		unmovable;
};


class Hair
{
public:
	int			m_height;
	int			m_width;
	int			m_segments;
	float		m_gap;
	float		m_lenGap;
	Point*		m_points;
	int			m_numPoints;
	Constraint* m_constraints;
	int			m_numConstraints;

	Hair()
	{
		m_width    = 15;
		m_height   = 15;
		m_segments = 10;
		m_gap	   = 0.3f;
		m_lenGap   = 2.5f;
		float distOffGround = 15.0f;

		m_numPoints = m_width * m_height * m_segments;
		m_points		= new Point[ m_numPoints ];

		m_numConstraints = m_width * m_height * (m_segments-1);
		m_constraints	= new Constraint[ m_numConstraints ];
		
		D3DXVECTOR3 bounds = D3DXVECTOR3(m_width*m_gap, 0, m_height*m_gap);
		D3DXVECTOR3 offset = bounds*0.5f + D3DXVECTOR3(0,0,-4);

		for (int x=0; x<m_width; x++)
		{
			for (int y=0; y<m_height; y++)
			{
				for (int s=0; s<m_segments; s++)
				{
					D3DXVECTOR3 pos = D3DXVECTOR3(x*m_gap, distOffGround - s*m_lenGap, y*m_gap) - offset;

					int indx = x + y*m_width + s*m_width*m_height;

					m_points[indx].curPos	 = pos;
					m_points[indx].oldPos	 = pos;
					m_points[indx].unmovable = false;
					
					if (s==0) m_points[indx].unmovable = true;
				}
			}
		}

		// Link together the constraints
		int i = 0;
		for (int s=0; s<m_segments-1; s++)
		{
			for (int x=0; x<m_width; x++)
			{
				for (int y=0; y<m_height; y++)
				{
					int indx0 = x + y*m_width + s*m_width*m_height;
					int indx1 = x + y*m_width + (s+1)*m_width*m_height;

					D3DXVECTOR3 p0 = m_points[indx0].curPos;
					D3DXVECTOR3 p1 = m_points[indx1].curPos;
					float len = D3DXVec3Length(&(p1-p0));

					m_constraints[i].index0		= indx0;
					m_constraints[i].index1		= indx1;
					m_constraints[i].restLength  = len;
					i++;
				}
			}

			if (i>m_numConstraints) DBG_HALT;
		}
		if (i!=m_numConstraints) DBG_HALT;


	}

	~Hair()
	{
	}

	void Shift(float x, float y, float z)
	{
		for (int i=0; i<m_numPoints; i++)
		{
			if (m_points[i].unmovable)
			{
				m_points[i].curPos.x += x;
				m_points[i].curPos.y += y;
				m_points[i].curPos.z += z;

				m_points[i].oldPos = m_points[i].curPos;
			}
		}
	}

	void VerletIntegrate(float dt)
	{
		for (int i=0; i<m_numPoints; i++)
		{
			if (m_points[i].unmovable) continue;

			D3DXVECTOR3 oldPos = m_points[i].oldPos;
			D3DXVECTOR3 curPos = m_points[i].curPos;
			D3DXVECTOR3 a = D3DXVECTOR3(0,-9.8f,0);

			curPos = 2*curPos - oldPos + a*dt*dt;

			m_points[i].oldPos = m_points[i].curPos;
			m_points[i].curPos = curPos;
		}
	}

	void SatisfyConstraints()
	{
		const int numIterations = 10;

		for (int i=0; i<numIterations; i++)
		{
			for (int k=0; k< m_numConstraints; k++)
			{

				// Constraint 2 (Cloth)
				Constraint* c = &m_constraints[k];
				D3DXVECTOR3& p0 = m_points[c->index0].curPos;
				D3DXVECTOR3& p1 = m_points[c->index1].curPos;
				D3DXVECTOR3 delta = p1-p0;
				float len = D3DXVec3Length(&delta);
				float diff = (len - c->restLength) / len;

				float unmove0 = m_points[c->index0].unmovable ? 0 : 1.0f;
				float unmove1 = m_points[c->index1].unmovable ? 0 : 1.0f;

				if (unmove0<=0.0f && unmove1<=0.0f) continue;

				float ratio = 1.0f/(unmove0+unmove1);

				p0 += delta*diff*(unmove0*ratio);
				p1 -= delta*diff*(unmove1*ratio);

				// Constraint 3 Other objects (Sphere)
				extern D3DXVECTOR3 g_headSphere; // Top Game.cpp
				extern float	   g_headRadius; // Top Game.cpp
				float radius	= g_headRadius * 2.2f;
				D3DXVECTOR3 dir = p0 - g_headSphere;
				float length	= D3DXVec3Length(&dir);
				if (length < radius)
				{
					D3DXVec3Normalize(&dir, &dir);
					p0 = g_headSphere + dir*radius;
				}

				dir = p1 - g_headSphere;
				length	= D3DXVec3Length(&dir);
				if (length < radius)
				{
					D3DXVec3Normalize(&dir, &dir);
					p1 = g_headSphere + dir*radius;
				}
			}
		}
	}

	void Update(float dt)
	{
		VerletIntegrate(dt);
		SatisfyConstraints();
	}

	void Render(IDirect3DDevice9* pDevice)
	{
		if (0) // If we want to draw little spheres for our points
		for (int i=0; i<m_numPoints; i++)
		{
			// Just for debug, draw a sphere
			static debugsphere ball(pDevice);
			ball.Draw(pDevice, 
					  m_points[i].curPos.x, 
					  m_points[i].curPos.y, 
					  m_points[i].curPos.z, 
					  0.2f);
		}

		// Draw a wire mesh for our constraints
		for (int i=0; i<m_numConstraints; i++)
		{
			Constraint* c = &m_constraints[i];
			DrawLine3D( pDevice, 
						m_points[c->index0].curPos, 
						m_points[c->index1].curPos,
						0xffff00ff);
		}
	}

};






#endif //HAIR_H


